"""HashMismatch is raised by the dispatcher system to indicate a hash
based mismatch between a type signature and a tuple of arguments. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.static import TypeSig

  from typing import Self


class HashMismatch(Exception):
  """HashMismatch is raised by the dispatcher system to indicate a hash
  based mismatch between a type signature and a tuple of arguments. """

  typeSig = _Attribute()
  posArgs = _Attribute()

  def __init__(self, typeSig: TypeSig, *args) -> None:
    """HashMismatch is raised by the dispatcher system to indicate a hash
    based mismatch between a type signature and a tuple of arguments. """

    self.typeSig = typeSig
    self.posArgs = args

    sigStr = str(typeSig)
    argTypes = [type(arg).__name__ for arg in args]
    argStr = """(%s)""" % ', '.join(argTypes)
    sigHash = hash(typeSig)
    if isinstance(args, list):
      argHash = hash((*args,))
    else:
      argHash = hash(args)

    infoSpec = """Unable to match type signature: <br><tab>%s<br>with
    signature of arguments:<br><tab>%s<br>Received hashes: %d != %d"""
    info = infoSpec % (sigStr, argStr, sigHash, argHash)
    Exception.__init__(self, info)

  def _resolveOther(self, other: object) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, TypeSig):
      return cls(other, *self.posArgs)
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Check if two HashMismatch objects are equal."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    if self.typeSig != other.typeSig:
      return False
    if self.posArgs != other.posArgs:
      return False
    return True



"""DataField represents an entry in the EZData classes. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False


class DataField:
  """DataField represents an entry in the EZData classes. """
  __slots__ = ('key', 'type_', 'val')

  def __init__(self, key: str, type_: type, val: object) -> None:
    """Initialize the DataField object."""
    self.key = key
    self.type_ = type_
    self.val = val



"""FidGen provides filename generator. Given a format specification and a
directory, it returns the next available filename of the given format."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.attr import WaitForIt, Field
from worktoy.mcls import BaseObject
from worktoy.parse import maybe
from worktoy.static import overload
from worktoy.text import stringList
from worktoy.waitaminute import TypeException, MissingVariable
from worktoy.work_io import validateExistingDirectory

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional


class FidGen(BaseObject):
  """
  FidGen provides filename generator. Given a format specification and a
  directory, it returns the next available filename of the given format.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  __field_name__ = None
  __field_owner__ = None
  __current_instance__ = None
  __current_owner__ = None
  fieldName = Field()
  fieldOwner = Field()
  currentInstance = Field()
  currentOwner = Field()

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: Any, owner: type) -> Any:
    """Get the value of the field."""
    self.currentInstance = instance
    self.currentOwner = owner
    return self

  def __set__(self, instance: Any, value: Any) -> None:
    """Set the value of the field."""
    self.currentInstance = instance
    self.baseName = value

  def __delete__(self, instance: Any) -> None:
    """Delete the field."""
    delattr(instance, self.fieldName)

  #  Fallback variables
  __fallback_extension__ = 'json'
  __fallback_directory__ = WaitForIt(os.getcwd)

  #  Private variables
  __base_name__ = None
  __file_extension__ = None
  __file_directory__ = None

  #  Public variables
  fileExtension = Field()
  fileDirectory = Field()
  baseName = Field()

  #  Virtual variables
  filePath = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  @fieldName.GET
  def _getFieldName(self) -> str:
    """Get the field name."""
    if self.__field_name__ is None:
      raise MissingVariable('__field_name__', str)
    return self.__field_name__

  @fieldOwner.GET
  def _getFieldOwner(self) -> type:
    """Get the field owner."""
    if self.__field_owner__ is None:
      raise MissingVariable('__field_owner__', type)
    return self.__field_owner__

  @currentInstance.GET
  def _getCurrentInstance(self) -> object:
    """Get the current instance."""
    return self.__current_instance__

  @currentOwner.GET
  def _getCurrentOwner(self) -> type:
    """Get the current owner."""
    return maybe(self.__current_owner__, self.__field_owner__, )

  @baseName.GET
  def _getFileName(self, ) -> str:
    """Get the name of the file."""
    return maybe(self.__base_name__, self.currentOwner.__name__)

  @fileExtension.GET
  def _getFileExtension(self, **kwargs) -> str:
    """Get the file extension."""
    return maybe(self.__file_extension__, self.__fallback_extension__, )

  @fileDirectory.GET
  def _getFileDirectory(self, **kwargs) -> str:
    """Get the file directory."""
    return maybe(self.__file_directory__, self.__fallback_directory__, )

  @filePath.GET
  def _getFilePath(self, **kwargs) -> str:
    """Getter-function for the file path. """
    n = kwargs.get('_n', 0)
    formatSpec = """%s_%03d"""
    fid = formatSpec % (self.baseName, n,)
    for item in os.listdir(self.fileDirectory):
      if item.startswith(fid):
        return self._getFilePath(_n=n + 1, )
    fidExt = """%s.%s""" % (fid, self.fileExtension)
    out = os.path.join(self.fileDirectory, fidExt)
    return str(out)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Setter methods
  @currentInstance.SET
  def _setCurrentInstance(self, instance: object) -> None:
    """Set the current instance."""
    if instance is not self.__current_instance__:
      self.__current_instance__ = instance

  @currentOwner.SET
  def _setCurrentOwner(self, owner: type) -> None:
    """Set the current owner."""
    if owner is not self.__current_owner__:
      self.__current_owner__ = owner

  @baseName.SET
  def _setFileName(self, newName: str) -> None:
    """Set the name of the file."""
    if self.__base_name__ == newName:
      return
    if not isinstance(newName, str):
      raise TypeException('__name__', newName, str)
    if not newName:
      raise ValueError('__name__ must be a non-empty string')
    self.__base_name__ = newName

  @fileExtension.SET
  def _setFileExtension(self, value: str) -> None:
    """Set the file extension."""
    if not isinstance(value, str):
      raise TypeException('__file_extension__', value, str)
    if not value:
      raise ValueError('__file_extension__ must be a non-empty string')
    self.__file_extension__ = value

  @fileDirectory.SET
  def _setFileDirectory(self, value: str, **kwargs) -> None:
    """Set the file directory."""
    if not isinstance(value, str):
      raise TypeException('__file_directory__', value, str)
    if not value:
      raise ValueError('__file_directory__ must be a non-empty string')
    try:
      validateExistingDirectory(value)
    except FileNotFoundError as e:
      if kwargs.get('_recursion', False):
        raise RecursionError from e
      os.makedirs(value, exist_ok=True)
      return self._setFileDirectory(value, _recursion=True)
    else:
      self.__file_directory__ = value
    finally:
      pass  # NOQA

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @staticmethod
  def _findDirectory(*args) -> Optional[str]:
    """Finds the directory in positional arguments. """
    for arg in args:
      if isinstance(arg, str):
        if os.path.isabs(arg):
          return arg
    return None  # pycharm, please!

  @staticmethod
  def _getCommonExtensions() -> list[str]:
    """Returns a list of common file extensions."""
    return stringList(
        """json, txt, csv, xml, html, pdf, doc, csv, py, 
        mkv, mp4, mp3, wav, jpg, png, gif, zip, tar, gz, bz2"""
    )

  @classmethod
  def _findFileExtension(cls, *args) -> Optional[str]:
    """Finds the file extension in positional arguments. """
    commonExtensions = cls._getCommonExtensions()
    for arg in args:
      if isinstance(arg, str):
        if str.startswith(arg, '*.'):
          return arg[2:]
        if arg in commonExtensions:
          return arg
    return None  # pycharm, please!

  @overload(str)
  def __init__(self, fileName: str, **kwargs) -> None:
    """Initialize the FidGen object."""
    self.__base_name__ = fileName
    self.__init__(**kwargs)

  @overload(str, str, str)
  @overload(str, str)
  def __init__(self, *args, **kwargs) -> None:
    argDir = self._findDirectory(*args)
    argExt = self._findFileExtension(*args)
    for arg in args:
      if arg in [argDir, argExt]:
        continue
      self.baseName = arg
      break
    if argDir is not None:
      self.fileDirectory = argDir
    if argExt is not None:
      self.fileExtension = argExt
    self.__init__(**kwargs)

  @overload()  # kwargs
  def __init__(self, **kwargs) -> None:
    """Initialize the FidGen object."""
    nameKeys = stringList("""name, file, fileName, filename, file_name""")
    extKeys = stringList("""ext, extension, file_extension""")
    dirKeys = stringList("""dir, directory, file_directory""")
    KEYS = [nameKeys, extKeys, dirKeys]
    NAMES = stringList("""name, ext, dirPath""")
    TYPES = dict(name=str, ext=str, dirPath=str)
    VALUES = dict()
    for (keys, (name, type_)) in zip(KEYS, TYPES.items()):
      for key in keys:
        if key in kwargs:
          value = kwargs[key]
          if isinstance(value, type_):
            VALUES[name] = value
            break
          raise TypeException(key, value, type_)



"""MetaNum provides the metaclass for the enumerations."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..attr import Field
from ..mcls import AbstractMetaclass, Base
from . import NumSpace as NSpace
from ..text import monoSpace, typeMsg
from ..waitaminute import UnrecognizedMember

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any, Never


class MetaNum(AbstractMetaclass):
  """MetaNum provides the metaclass for the enumerations."""

  __iter_contents__ = None

  memberType = Field()
  NULL = Field()

  @memberType.GET
  def _getMemberType(cls) -> type:
    """Get the member type."""
    type_ = getattr(cls, '__member_type__', None)
    if type_ is None:
      raise AttributeError('Member type not set!')
    if isinstance(type_, type):
      return type_
    raise TypeError(typeMsg('__member_type__', type_, type))

  @NULL.GET
  def _getNULL(cls, ) -> Any:
    """Get the NULL value."""
    nullMember = getattr(cls, '__null_member__', None)
    if nullMember is None:
      e = """KeeNum class: '%s' does not have a NULL member!"""
      info = monoSpace(e % cls.__name__)
      raise AttributeError(info)
    return nullMember

  @classmethod
  def __prepare__(mcls, name: str, bases: Base, **kwargs) -> NSpace:
    """Prepare the class namespace."""
    return NSpace(mcls, name, bases, **kwargs)

  def __new__(mcls, name: str, bases: Base, space: NSpace, **kwargs) -> type:
    """Create a new class."""
    return AbstractMetaclass.__new__(mcls, name, bases, space, **kwargs)

  def __init__(cls, name: str, bases: Base, space: NSpace, **kwargs) -> None:
    """The __init__ method is invoked to initialize the class."""
    AbstractMetaclass.__init__(cls, name, bases, space, **kwargs)
    memberType = None
    memberObjects = []
    for i, member in enumerate(space.getMemberNums()):
      if memberType is None:
        memberType = type(member.val)
      if not isinstance(member.val, memberType):
        e = """Inconsistent types of inner values in the enumeration! 
        Expected: type: '%s', but received object: '%s' of type: '%s'!"""
        expType = memberType.__name__
        objStr = str(member.val)
        actType = type(member.val).__name__
        info = monoSpace(e % (expType, objStr, actType))
        raise TypeError(info)
      existing = getattr(cls, '__member_objects__', )
      newMember = cls(member.key, member.val, i)
      setattr(cls, member.key, newMember)
      memberObjects.append(newMember)
      if member.key == 'NULL':
        if getattr(cls, '__null_member__', None) is not None:
          e = """KeeNum class: '%s' has multiple NULL members!"""
          info = monoSpace(e % cls.__name__)
          raise AttributeError(info)
        if member.val:
          e = """The NULL member must evaluate to False, but received: '%s'
          which evaluates to True!"""
          raise ValueError(monoSpace(e % member.val))
        setattr(cls, '__null_member__', newMember)
    else:
      setattr(cls, '__allow_instantiation__', False)
      setattr(cls, '__member_type__', memberType)
      setattr(cls, '__member_objects__', memberObjects)

  def __call__(cls, *args: Any, **kwargs: Any) -> Self:
    """The __call__ method is invoked to create an instance of the
    class."""
    if getattr(cls, '__allow_instantiation__', False):
      return AbstractMetaclass.__call__(cls, *args, **kwargs)
    return cls._resolveMember(*args, **kwargs)

  def _resolveMember(cls, *args, **kwargs) -> Any:
    """The _resolveMember method is invoked to resolve the member
    variables."""
    if len(args) != 1:
      e = """Cannot resolve member with more than one argument!"""
      raise ValueError(monoSpace(e))
    identifier = args[0]
    if isinstance(identifier, cls):
      return identifier
    if isinstance(identifier, int):
      try:
        return cls._resolveIndex(identifier)
      except IndexError as indexError:
        raise UnrecognizedMember(cls, identifier) from indexError
    if isinstance(identifier, str):
      try:
        return cls._resolveKey(identifier)
      except KeyError as keyError:
        raise UnrecognizedMember(cls, identifier) from keyError
    if isinstance(identifier, cls.memberType):
      for self in cls:
        if identifier == self.val:
          return self
    raise UnrecognizedMember(cls, identifier)

  def _resolveIndex(cls, index: int) -> Any:
    """The _resolveIndex method is invoked to resolve the index of the
    member variables."""
    while index < 0:
      index += len(cls)
    if index >= len(cls):
      e = """Index out of range!"""
      raise IndexError(monoSpace(e))
    for self in cls:
      if self.index == index:
        return self

  def _resolveKey(cls, key: str) -> Any:
    """The _resolveKey method is invoked to resolve the key of the member
    variables."""
    for self in cls:
      if self.key.lower() == key.lower():
        return self
    e = """Key not found!"""
    raise KeyError(monoSpace(e))

  def __iter__(cls, ) -> Self:
    """The __iter__ method is invoked to iterate over the class."""
    cls.__iter_contents__ = getattr(cls, '__member_objects__', )
    return cls

  def __next__(cls) -> Any:
    """The __next__ method is invoked to get the next item in the
    iteration."""
    if cls.__iter_contents__:
      return cls.__iter_contents__.pop(0)
    raise StopIteration

  def __len__(cls, ) -> int:
    """The __len__ method is invoked to get the length of the class."""
    return len(getattr(cls, '__member_objects__', []))

  def __bool__(cls, ) -> bool:
    """The __bool__ method is invoked to get the boolean value of the
    class."""
    return True if getattr(cls, '__member_objects__', []) else False

  def __contains__(cls, item: Any) -> bool:
    """The __contains__ method is invoked to check if the class contains
    the item."""
    if item in getattr(cls, '__member_objects__', []):
      return True
    if item in cls.memberType:
      for self in cls:
        if self == item:
          return True
    return False

  def __getitem__(cls, key: str) -> Any:
    """The __getitem__ method is invoked to get the item in the class."""
    return cls._resolveKey(key)

  def __setitem__(cls, *_) -> Never:
    """Illegal operation"""
    info = """KeeNum classes are immutable, but an attempt was 
    made to set an item on KeeNum class: :'%s'!""" % cls.__name__
    raise TypeError(monoSpace(info))

  def __delitem__(cls, *_) -> Never:
    """Illegal operation"""
    info = """KeeNum classes are immutable, but an attempt was 
    made to delete an item on KeeNum class: :'%s'!""" % cls.__name__
    raise TypeError(monoSpace(info))

  def __getattr__(cls, key: str) -> Any:
    """The __getattr__ method is invoked to get the attribute of the
    class."""
    try:
      return cls._resolveKey(key)
    except KeyError as keyError:
      try:
        return object.__getattribute__(cls, key)
      except AttributeError as attributeError:
        raise attributeError from keyError



"""EZHook collects the field entries in EZData class bodies. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..mcls import FunctionType
from ..mcls.hooks import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable
  from worktoy.ezdata import EZSpace, EZData


class _EZIter:
  """_EZIter provides an iterator for the EZData class. """

  __iter_contents__ = None

  def __init__(self, *args) -> None:
    """The __init__ method is called when the class is created."""
    self.__iter_contents__ = [*args, ]

  def __next__(self, ) -> Any:
    """The __next__ method is called when the class is created."""
    if self.__iter_contents__:
      return self.__iter_contents__.pop(0)
    self.__iter_contents__ = None
    raise StopIteration()


class EZHook(AbstractHook):
  """EZHook collects the field entries in EZData class bodies. """

  @staticmethod
  def initFactory(*dataFields) -> FunctionType:
    """The initFactory method is called when the class is created."""

    slotKeys = [dataField.key for dataField in dataFields]
    defVals = [dataField.val for dataField in dataFields]

    def __init__(self, *args, **kwargs):
      """The __init__ method is called when the class is created."""
      posArgs = [*args, ]
      while len(posArgs) < len(slotKeys):
        posArgs.append(None)
      for key, defVal in zip(slotKeys, defVals):
        setattr(self, key, defVal)
      for key, arg in zip(slotKeys, posArgs):
        if arg is not None:
          setattr(self, key, arg)
      for key in slotKeys:
        if key in kwargs:
          setattr(self, key, kwargs[key])

    return __init__

  @staticmethod
  def eqFactory(*dataFields) -> FunctionType:
    """The eqFactory method is called when the class is created."""

    def __eq__(self, other: EZData) -> bool:
      """The __eq__ method is called when the class is created."""
      for dataField in dataFields:
        key = dataField.key
        if getattr(self, key) != getattr(other, key):
          return False
      return True

    return __eq__

  @staticmethod
  def strFactory(*dataFields) -> FunctionType:
    """The strFactory method is called when the class is created."""

    def __str__(self) -> str:
      """The __str__ method is called when the class is created."""
      clsName = type(self).__name__
      keys = [dataField.key for dataField in dataFields]
      vals = [str(getattr(self, key)) for key in keys]
      keyVals = ['%s=%s' % (key, val) for key, val in zip(keys, vals)]
      return """%s(%s)""" % (clsName, ', '.join(keyVals))

    return __str__

  @staticmethod
  def reprFactory(*dataFields) -> FunctionType:
    """The reprFactory method is called when the class is created."""

    def __repr__(self) -> str:
      """The __repr__ method is called when the class is created."""
      clsName = type(self).__name__
      keys = [dataField.key for dataField in dataFields]
      vals = [str(getattr(self, key)) for key in keys]
      return """%s(%s)""" % (clsName, ', '.join(vals))

    return __repr__

  @staticmethod
  def iterFactory(*dataFields) -> FunctionType:
    """The iterFactory method is called when the class is created."""

    def __iter__(self, ) -> _EZIter:
      """The __iter__ method is called when the class is created."""
      keys = [dataField.key for dataField in dataFields]
      vals = [getattr(self, key) for key in keys]
      return _EZIter(*vals)

    return __iter__

  @staticmethod
  def getItemFactory(*dataFields) -> FunctionType:
    """The getItemFactory method is called when the class is created."""

    def __getitem__(self, key: str) -> object:
      """The __getitem__ method is called when the class is created."""
      if key in self.__slots__:
        return getattr(self, key)
      raise KeyError(key)

    return __getitem__

  @staticmethod
  def setItemFactory(*dataFields) -> FunctionType:
    """The setItemFactory method is called when the class is created."""

    def __setitem__(self, key: str, value: object) -> None:
      """The __setitem__ method is called when the class is created."""
      if key in self.__slots__:
        return setattr(self, key, value)
      raise KeyError(key)

    return __setitem__

  def setItemHook(
      self,
      space: EZSpace,
      key: str,
      value: object,
      oldValue: object
  ) -> bool:
    """The setItemHook method is called when an item is set in the
    enumeration."""
    if callable(value):
      return False
    space.addField(key, type(value), value)
    return True

  def preCompileHook(self, space: EZSpace, namespace: dict) -> dict:
    """The preCompileHook method is called before the class is compiled."""
    dataFields = space.getDataFields()
    namespace['__slots__'] = (*[dataField.key for dataField in dataFields],)
    return namespace

  def postCompileHook(self, space: EZSpace, namespace: dict) -> dict:
    """The postCompileHook method is called after the class is compiled."""
    dataFields = space.getDataFields()
    initMethod = self.initFactory(*dataFields)
    eqMethod = self.eqFactory(*dataFields)
    strMethod = self.strFactory(*dataFields)
    reprMethod = self.reprFactory(*dataFields)
    iterMethod = self.iterFactory(*dataFields)
    getItemMethod = self.getItemFactory(*dataFields)
    setItemMethod = self.setItemFactory(*dataFields)
    namespace['__init__'] = initMethod
    namespace['__eq__'] = eqMethod
    namespace['__str__'] = strMethod
    namespace['__repr__'] = reprMethod
    namespace['__iter__'] = iterMethod
    namespace['__getitem__'] = getItemMethod
    namespace['__setitem__'] = setItemMethod
    return namespace



"""TypeException is a custom exception class for handling type related
errors. Specifically, this exception should NOT be raised if the object is
None instead of the expected type. This is because None indicates absense
rather than type mismatch. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from warnings import warn

from . import _Attribute

from ..text import monoSpace, joinWords

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self, Callable


def _resolveTypeNames(*types) -> str:
  """Creates the first part of the error message listing the expected type
  or types. """
  if len(types) == 1:
    if isinstance(types[0], (tuple, list)):
      return _resolveTypeNames(*types[0])
    if isinstance(types[0], type):
      expName = types[0].__name__
    elif isinstance(types[0], str):
      expName = types[0]
    else:
      raise TypeError("""Received bad arguments: %s""" % (str(types),))
    return """Expected object of type '%s'""" % (expName,)
  typeNames = []
  for type_ in types:
    if isinstance(type_, type):
      typeNames.append("""'%s'""" % type_.__name__)
    elif isinstance(type_, str):
      typeNames.append("""'%s'""" % type_)
    else:
      raise TypeError("""Received bad arguments: %s""" % (str(types),))
  infoSpec = """Expected object of any of the following types: %s"""
  typeStr = joinWords(*typeNames, sep='or')
  return monoSpace(infoSpec % (typeStr,))


class TypeException(TypeError):
  """
  TypeException is a custom exception class for handling type related
  errors. Specifically, this exception should NOT be raised if the object is
  None instead of the expected type.
  """

  varName = _Attribute()
  actType = _Attribute()
  expTypes = _Attribute()

  def __init__(self, name: str, obj: object, *types) -> None:
    """Initialize the TypeException with the name of the variable, the
    received object, and the expected types."""
    prelude = _resolveTypeNames(*types)
    actName = type(obj).__name__
    infoSpec = """%s at name: '%s', but received object of type '%s' with 
    repr: '%s'"""
    info = infoSpec % (prelude, name, actName, repr(obj))
    TypeError.__init__(self, monoSpace(info))
    self.varName = name
    self.actType = type(obj)
    self.expTypes = types

  def _resolveOther(self, other: Any) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the TypeException object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)
    if isinstance(other, cls):
      if self.varName != other.varName:
        return False
      if self.actType != other.actType:
        return False
      if self.expTypes != other.expTypes:
        return False
      return True
    return False



"""TypeSig instances represent particular type signatures. The 'fast' method
provides a performant validation of a tuple of objects. The 'flex' method
attempts to cast each object in a tuple to the expected type. The latter is
substantially slower than the former."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import typeMsg
from ..waitaminute import CastMismatch, HashMismatch
from .casting import AbstractCast, Cast
from . import THIS

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, TypeAlias, Never

  Types: TypeAlias = list[type]
  Casts: TypeAlias = tuple[AbstractCast, ...]
  TypeCasts: TypeAlias = dict[type, AbstractCast]


class TypeSig:
  """TypeSig instances represent particular type signatures. The 'fast'
  method provides a performant validation of a tuple of objects. The
  'flex' method attempts to cast each object in a tuple to the expected
  type. The latter is substantially slower than the former."""

  __iter_contents__ = None
  __raw_types__ = None
  __type_casts__ = None
  __hash_value__ = None

  def __init__(self, *types: Any) -> None:
    """Initialize the TypeSig instance."""
    if len(types) == 1:
      if isinstance(types[0], tuple):
        types = types[0]
    self.__raw_types__ = types

  def getCasts(self, ) -> Casts:
    """Get the casts for the types."""
    out = []
    for type_ in self.__raw_types__:
      if type_ is THIS:
        raise RuntimeError('THIS not yet replaced!')
      out.append(Cast(type_))
    return (*out,)

  def replaceTHIS(self, cls: type) -> None:
    """Replace the 'THIS' type with the class."""
    newTypes = (*[cls if t is THIS else t for t in self.__raw_types__],)
    self.__raw_types__ = newTypes
    self.__hash_value__ = hash(newTypes)

  def __hash__(self, ) -> int:
    """Forwards the hash to the hash of the types. If the replaceTHIS
    method has not been called, this will raise RuntimeError."""
    if self.__hash_value__ is None:
      return hash(self.__raw_types__)
    return self.__hash_value__

  def __eq__(self, other: object) -> bool:
    """Check if the other object is a TypeSig and has the same types."""
    cls = type(self)
    if not isinstance(other, cls):
      return self == cls(other)
    return True if hash(self) == hash(other) else False

  def __len__(self, ) -> int:
    """Get the length of the types."""
    return len(self.__raw_types__)

  def __str__(self, ) -> str:
    """String representation reflects the types. """
    typeStr = ', '.join([t.__name__ for t in self.__raw_types__])
    info = """%s: [%s]""" % (type(self).__name__, typeStr)
    return info

  def __repr__(self, ) -> str:
    """Code representation reflects the types. """
    typeStr = ', '.join([t.__name__ for t in self.__raw_types__])
    info = """%s(%s)""" % (type(self).__name__, typeStr)
    return info

  def fast(self, *args) -> tuple:
    """Check if the types of the arguments match the types in the
    signature."""
    if len(args) == len(self):
      if hash(self) == hash((*[type(arg) for arg in args],)):
        return (*args,)
    raise HashMismatch(self, *args)

  def flex(self, *args) -> tuple:
    """Cast the arguments to the types in the signature."""
    if len(args) != len(self):
      if len(args) == 1:
        if isinstance(args[0], (list, tuple)):
          return self.fast(*args[0])
      raise HashMismatch(self, *args)
    casts = self.getCasts()
    types = self.__raw_types__
    out = []
    for arg, type_, cast in zip(args, types, casts):
      if not isinstance(type_, type):
        raise TypeError(typeMsg('type_', type_, type))
      try:
        if isinstance(arg, type_):
          out.append(arg)
        else:
          out.append(cast(arg))
      except CastMismatch:
        continue
      except RecursionError:
        continue
    return (*out,)



"""BaseSpace provides the namespace class for the worktoy.mcls module."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import VariableNotNone
from . import AbstractNamespace
from .hooks import AccessorHook, OverloadHook, NameHook, ReservedNameHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import TypeAlias, Callable, Any
  from worktoy.static import TypeSig
  from worktoy.mcls import FunctionType as Func

  OverloadMap: TypeAlias = dict[str, dict[TypeSig, Callable[..., Any]]]


class BaseSpace(AbstractNamespace):
  """BaseSpace provides the namespace class for the worktoy.mcls module."""

  __overload_map__ = None

  def _buildOverloadMap(self, ) -> None:
    """Build the overload map for the namespace."""
    if self.__overload_map__ is not None:
      raise VariableNotNone('__overload_map__', )
    if self.isPrime():
      self.__overload_map__ = {}
      return
    parentSpace = self.getParentSpace()
    self.__overload_map__ = {**parentSpace.getOverloadMap(), }

  def getOverloadMap(self, **kwargs) -> OverloadMap:
    """Get the overload map for the namespace."""
    if self.__overload_map__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._buildOverloadMap()
      return self.getOverloadMap(_recursion=True)
    return self.__overload_map__

  def addOverload(self, key: str, sig: TypeSig, func: Func) -> None:
    """Add an overload to the overload map."""
    overloadMap = self.getOverloadMap()
    existingMap = overloadMap.get(key, {})
    existingMap[sig] = func
    overloadMap[key] = existingMap
    self.__overload_map__ = overloadMap

  reservedNameHook = ReservedNameHook()
  accessorHook = AccessorHook()
  overloadHook = OverloadHook()
  nameHook = NameHook()



"""EZMeta provides the metaclass for the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..mcls import AbstractMetaclass, Base
from ..ezdata import EZSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class _WhenReady:
  """WhenReady is a decorator that is used to mark a method as ready
  for use. """

  name = None
  bases = None
  space = None
  kwargs = None

  __wrapped__ = None

  def __init__(self, bases: tuple, **kwargs) -> None:
    """Initialize the WhenReady object."""
    self.bases = bases
    self.kwargs = kwargs

  def __set_name__(self, owner: type, name: str) -> None:
    self.name = name
    self.owner = owner
    space = EZSpace(EZMeta, name, self.bases, )
    for key, val in self.kwargs.items():
      space[key] = val
    space['__name__'] = name
    self.__wrapped__ = EZMeta(self.name, self.bases, space)

  def __get__(self, instance: object, owner: type) -> EZMeta:
    """Get the WhenReady object."""
    return self.__wrapped__


class EZMeta(AbstractMetaclass):
  """EZMeta provides the metaclass for the EZData class."""

  @classmethod
  def __prepare__(mcls, name: str, bases: Base, **kwargs: dict) -> EZSpace:
    """Prepare the class namespace."""
    return EZSpace(mcls, name, bases, **kwargs)

  def __call__(cls, *args: tuple, **kwargs: dict) -> Any:
    """Call the class."""
    if cls.__name__ == 'EZData':
      return _WhenReady((cls,), **kwargs)
    return AbstractMetaclass.__call__(cls, *args, **kwargs)



"""ReservedNameHook protects reserved names. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import ReservedName
from . import AbstractHook, ReservedNames

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.mcls import AbstractNamespace as ASpace


class ReservedNameHook(AbstractHook):
  """ReservedNameHook protects reserved names."""

  reservedNames = ReservedNames()

  def setItemHook(
      self,
      space: ASpace,
      key: str,
      value: object,
      oldValue: object
  ) -> bool:
    """The setItemHook method is called when an item is set in the
    namespace."""
    if key in self.reservedNames:
      if key not in space:
        dict.__setitem__(space, key, value)
        return True
      raise ReservedName(key)
    return False
